;***************************************************************************
;*
;* Title	: AVR ISP (Auto adr inc, 19200-115200bps)
;* Version	: 4.0
;* Last updated	: June 26 2004 (plinortner)
;* Target	: AT90S1200/AT90S2313/...
;* File		: avr910.asm (requires AVRStudio 4.0 or higher)
;* Author(s)	: Ole Saether, Terje Frostad,
;*		  Ingar Fredriksen, Morten W. Lund
;*                Haakon Skar, Paal Kastnes,
;*		  Peter Linortner
;* 
;* DESCRIPTION
;*	The firmware on all programmers now support a unified protocol for 
;*	program and data memory programming. The host computer do not need
;*	to know if the programmer operates in serial or parallel mode.
;*
;*	The following commands are supported. All commands start with a
;*	single letter. The programmer returns 13d (carriage return) or the
;*	data read after the command is finished.
;*
;*                                     +-------------+------------+------+
;*  Commands                           | Host writes | Host reads |      |
;*  --------                           +-----+-------+------+-----+      |
;*                                     | ID  | data  | data |     | Note |
;* +-----------------------------------+-----+-------+------+-----+------+
;* |BASIC:                             |     |       |      |     |      |
;* | Enter programming mode            | 'P' |       |      | 13d |   1  |
;* | Report autoincrement address      | 'a' |       | 	    | 'Y' |      |
;* | Set address                       | 'A' | ah al |      | 13d |   2  |
;* | Write program memory, low byte    | 'c' |    dd |      | 13d |   3  |
;* | Write program memory, high byte   | 'C' |    dd |      | 13d |   3  |
;* | Issue Page Write                  | 'm' |       |      | 13d |      |
;* | Read program memory               | 'R' |       |dd dd |     |   4  |
;* | Write data memory                 | 'D' |    dd |      | 13d |      |
;* | Read data memory                  | 'd' |       |   dd |     |      |
;* | Chip erase                        | 'e' |       |      | 13d |      |
;* | Write lock bits                   | 'l' |    dd |      | 13d |      |
;* | New universal command	       | '.' |  4*dd |   dd | 13d |      |
;* | Leave programming mode            | 'L' |       |      | 13d |   5  |
;* | Select device type                | 'T' |    dd |      | 13d |   6  |
;* | Read signature bytes              | 's' |       | 3*dd |     |      |
;* | Return supported device codes     | 't' |       | n*dd | 00d |   7  |
;* | Return software identifier        | 'S' |       | s[7] |     |   8  |
;* | Return sofware version            | 'V' |       |dd dd |     |   9  |
;* | Return hardware version           | 'v' |       |dd dd |     |   9  |
;* | Return programmer type            | 'p' |       |   dd |     |  10  |
;* | Set LED                           | 'x' |    dd |      | 13d |  12  |
;* | Clear LED                         | 'y' |    dd |      | 13d |  12  |
;* +-----------------------------------+-----+-------+------+-----+------+
;* |NEW:                               |     |       |      |     |      |
;* | Report block write support        | 'b' |       |'Y'2*dd| 13d|  14  |
;* | Write block to memory             | 'B' | 3*dd  |      | 13d |  14  |
;* | Read block from memory            | 'g' | 3*dd  | 3*dd | 13d |  14  |
;* | Exit upgrade                      | 'L' |       |      | 13d |   5  |
;* +-----------------------------------+-----+-------+------+-----+------+
;* |NOT SUPPORTED:                     |     |       |      |     |      |
;* | Universial command                | ':' |  3*dd |   dd | 13d |  16  |
;* | Return lock bits                  | 'r' |       |   dd | 13d |  15  |
;* | Return high fuse bits             | 'N' |       |   dd | 13d |  15  |
;* | Return extended fuse bits         | 'Q' |       |   dd | 13d |  15  |
;* | Write fuse bits                   | 'f' |    dd |      | 13d |  11  |
;* | Read fuse and lock bits           | 'F' |       |   dd |     |  11  |
;* | Special test command	       | 'Z' |  2*dd |   dd |     |      |
;* +-----------------------------------+-----+-------+------+-----+------+
;*
;* NOTE 1
;*	The Enter programming mode command MUST be sent one time prior to
;*	the other commands, with the exception of the 't', 'S', 'V', 'v'
;*	and 'T' commands. The 'T' command must be sent before this command
;*	(see note 6).
;*
;*	For programmers supporting both parallel and serial programming
;*	mode this command enters parallel programming mode. For programmers
;*	supporting only serial programming mode, this command enters serial
;*	programming mode.
;*
;* NOTE 2
;*	The ah and al are the high and low order bytes of the address. For
;*	parallel programmers this command issues the Load Address Low/High
;*	Byte command. For serial programmers the address byte is stored for
;*	use by the Read/Write commands.
;*
;* NOTE 3
;*	For parallel programmers this command issues the Program Flash
;*	command. For serial programmers this command iussues the Write
;*	Program Memory Command. For devices with byte-wide program memories
;*	only the low byte command should be used.
;*
;* NOTE 4
;*	The contents of the program memory at the address given by the 'A'
;*	command are written to the serial port in binary form. For byte
;*	wide memories one byte is written. For 16 bit memories two bytes
;*	are written,MSB first.
;*
;* NOTE 5
;*	This command must be executed after the programming is finished.
;*
;* NOTE 6
;*	The select device type command must be sent before the enter
;*	programming command
;*
;* NOTE 7
;*	The supported device codes are returned in binary form terminated
;*	by 0x00.
;*
;* NOTE 8
;*	This return a 7 character ASCII string identifying the programmer.
;*	For the development board it is "AVR DEV", for the parallel
;*	programmer it is "AVR PPR" and for the in-curcuit programmer it is
;*	"AVR ICP".
;*
;* NOTE 9
;*	The software/hardware version are returned as two ASCII numbers.
;*
;* NOTE 10
;*	This command should be used to identify the programmer type. The
;*	return value is 'S' for serial (or SPI) programmers or 'P' for
;*	parallel programmers.
;*
;* NOTE 11
;*	The write fuse bits command are available only on parallel
;*	programmers and only for AVR devices (device code < 0x80). The host
;*	should use the return programmer type command to determine the
;*	programmer type, do not use the  "AVR PPR" idenifier because other
;*	programmers may be available in the future.
;*
;* NOTE 12
;*	Currently only the AVR development board has LEDs. The other boards
;*	must implement this commands as NOPs.
;*
;* NOTE 13
;*      Devices using Page Mode Programming write one page of flash memory
;*      before issuing a Page Mode Write Pulse.
;*
;* NOTE 14
;*	Commands added for bandwith increase
;*
;* NOTE 15
;*	Unsupported (can be emulated with 'new universal command')
;*
;* NOTE 16
;*	Unsupported (only required for AT89Sxxx devices)
;*
;*
;*
;* HISTORY				
;*	V4.0    04.25.06 (plinortner)	Added AVRProg compatible block write commands
;*                                      Added polling for flash programming 
;*                                      (taken from V3.2)
;*                                      Added support for multible targets (1200/2313/...)
;*                                      Relaxed SPI timing
;*                                      Added (optional) LED support
;*                                      Added hardware/software UART option
;*      V3.0    04.11.06 (plinortner) 	Decreased code size to less than 512 words
;*					Added support for multiple new devices
;*					Changed software UART code to support 115200 bps
;*      V2.2    00.03.10 (pkastnes)     Added support for multiple new devices
;*	V2.1	98.10.26 (mlund)	New date marking.
;*					Removed support for AT90S1200C.
;*					Added support for AT90S4433A.
;*	V2.0	98.01.06 (mlund)	ATmega103 support
;*	V1.7	97.11.06 (mlund)	Universial command (':') implemented.
;*					Releases all pins when not in 
;*					programming mode.
;*	V1.6e	97.11.04 (mlund)	mega103 rev D support
;*	V1.6c	97.10.30 (mlund)	Auto incrementing / SPI sync
;*					also works for mega103.
;*      V1.6	97.09.09 (hskar)	Created Page Mode Version (mega103)
;*	V1.5	97.08.21 (mlund)	Modified / Bugfix / Major cleanup
;*	...	...			(no records)
;*	V?.?	97.03.15 (OS)		Created
;* 
;* 
;*		
;* Device Support List    
;* -------------------
;* 
;* +----------+----------+------+-------+------+------+------+-------+
;* |Device    |Signature | Code | Flash |EEProm| Lock | Fuse | PMode |
;* +----------+----------+------+-------+------+------+------+-------+
;* | tiny10   |          | 0x51 |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* | tiny11   |          | 0x50 |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* | tiny12   | 1E 90 05 | 0x55 |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* | tiny15   | 1E 90 06 | 0x56 |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* | tiny19   |          | 0x58 |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* | tiny26   |          | 0x5E |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* | tiny28   |          | 0x5C |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* |          |          |      |       |      |      |      |       |
;* | S1200    | 1E 90 01 | 0x13 |  R/W  | R/W  |  W   | NA   | Byte  |
;* |          |          |      |       |      |      |      |       |
;* | S2313    | 1E 91 01 | 0x20 |  R/W  | R/W  |  W   | NA   | Byte  |
;* | S2323    | 1E 91 02 | 0x48 |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* | S2333    | 1E 91 05 | 0x34 |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* | S2343    | 1E 91 03 | 0x4C |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* |          |          |      |       |      |      |      |       |        
;* | mega8515 |          | 0x3A |  R/W  | R/W  | R/W  | R/W  | Page  |
;* | mega8515B|          | 0x3A |  R/W  | R/W  | R/W  | R/W  | Page  |
;* | mega8535 |          | ???? |  R/W  | R/W  | R/W  | R/W  | Page  |
;* | mega8535B|          | ???? |  R/W  | R/W  | R/W  | R/W  | Page  |
;* |          |          |      |       |      |      |      |       |
;* | mega8    |          | 0x76 |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega8B   |          | 0x77 |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega16   |          | 0x74 |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega16B  |          | 0x75 |  R/W  | R/W  | R/W  | R/W  | Page  |    
;* | mega32   |          | 0x72 |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega32B  |          | 0x73 |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega64   |          | ???? |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega64B  |          | ???? |  R/W  | R/W  | R/W  | R/W  | Page  |     
;* | mega128  |          | 0x43 |  R/W  | R/W  | R/W  | R/W  | Page  |   
;* | mega128B |          | 0x44 |  R/W  | R/W  | R/W  | R/W  | Page  |
;* +----------+----------+------+-------+------+------+------+-------+
;* | S4414    | 1E 92 01 | 0x28 |  R/W  | R/W  |  W   | NA   | Byte  |    
;* | S4433    | 1E 92 03 | 0x30 |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* | S4434    | 1E 92 02 | 0x6C |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* |          |          |      |       |      |      |      |       |
;* | S8515    | 1E 93 01 | 0x38 |  R/W  | R/W  |  W   | NA   | Byte  |    
;* | S8535    | 1E 93 03 | 0x68 |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* |          |          |      |       |      |      |      |       |
;* | C8534    |	         | 0x70 |  R/W  | R/W  | R/W  | R/W  | Byte  |    
;* | C8544    |          | 0x71 |  R/W  | R/W  | R/W  | R/W  | Byte  |
;* |          |          |      |       |      |      |      |       |    
;* | mega83   | 1E 93 05 | 0x65 |  R/W  | R/W  | R/W  | R/W  | Page  |    
;* | mega83B  |          | 0x67 |  R/W  | R/W  | R/W  | R/W  | Page  |        
;* | mega161  | 1E 94 01 | 0x60 |  R/W  | R/W  | R/W  | R/W  | Page  |    
;* | mega161B |          | 0x61 |  R/W  | R/W  | R/W  | R/W  | Page  |    
;* | mega163  | 1E 94 02 | 0x64 |  R/W  | R/W  | R/W  | R/W  | Page  |
;* | mega163B |          | 0x66 |  R/W  | R/W  | R/W  | R/W  | Page  |
;* | mega103  | 1E 97 01 | 0x41 |  R/W  | R/W  | R/W  | R/W  | Page  |
;* +----------+----------+------+-------+------+------+------+-------+
;* 
;* LEGEND:
;* -------
;* Signature - Device Signature Byte
;* Code      - Unique device code used by AVRProg to identify device
;* Flash     - Flash access type
;* EEProm    - EEProm access type
;* Lock      - Lockbits access type
;* Fuse      - Fusebits access type
;* PMode     - Indicates if device uses byte or page programming mode
;* 
;* R/W - Read and Write Access
;* R   - Read Access Only
;* W   - Write Access Only
;* NA  - Not Accessible
;* 
;***************************************************************************

;boolean constants
.equ	TRUE  = 1
.equ	FALSE = 0

;*****************************
;* Firmware Configuration    *
;*****************************

.include "configure.inc"


;*****************************
;* Protocoll Definitions     *
;*****************************

.include "commands.inc"

;**** Check User Inputs ****

.if SUPPORT_BLOCK_CMD==FALSE && SUPPORT_BYTE_CMD==FALSE
.error "Error: Block commands and/or bye commands must be enabled in order to program a device"
.endif

;**** Baud Rate Calculation ****

;+--------+--------+---+---+-----------+
;| BAUD   | BITLEN | R | C | XTAL      |
;+--------+--------+---+---+-----------|
;| 115200 | 35     | 1 | 1 | 4.000 MHz |
;|  38400 | 104    | 1 | 1 | 4.000 MHz |
;|  19200 | 26     | 2 | 8 | 4.000 MHz |
;| ...    | ...    |...|...| ...       |
;+--------+--------+---+---+-----------+

.equ	BITLEN = (CLK+UART_BAUD/2) / UART_BAUD
.equ	R = 1 * (BITLEN<=170) + 2 * (BITLEN>170)
.equ	C = 1 * (R     ==1  ) + 8 * (R==2)
.equ	N = BITLEN/C

;**** Buffer Configuration ****

.ifdef 	SPL
.message "Configuration: use SRAM as RX buffer area"
.equ	USE_RAM=TRUE
.else
.message "Configuration: use register file as RX buffer area"
.equ	USE_RAM=FALSE
.if	UART_RXSIZE>16
.error	"Error: RX buffer cannot be greater than 16 bytes for this device"
.endif 
.endif

;**** Stack Configuration ****

.ifdef	SPL
.equ	STACKTOP	= RAMEND
.endif

;*****************************
;* Global Register Variables *
;*****************************

.if	USE_RAM==TRUE
.equ	BUFEND	        = RAMEND-8; UART Rx Buffer
.def	count		= r15	; Counter for RAM devices
.else
;def	rx_buffer	= r0-r15; UART Rx Buffer
.endif
;Polling
.def	p_cmd		= r16	; Flash data polling
.def	p_addrh		= r17	; Flash data polling
.def	p_addrl		= r18	; Flash data polling
.def	p_data		= r19	; Flash data polling
;UART
.def	u_data		= r20	; UART data
;SPI
.def	s_data		= r21	; SPI  data 
;working (device drivers)
.def	temp1		= r22	; Temporary Register 1
.def	temp2		= r23	; Temporary Register 2
;working (user interface)
.def	cmd1		= r24	; Working register 1
.def	cmd2		= r25	; Working register 2
.def	cmd3		= r26	; Working register 3
;external interface
.def	device		= r27	; Device code
.def	addrl		= r28	; Low order byte of address
.def	addrh		= r29	; High order byte of address

.CSEG

;*********************
;* Interrupt Vectors *
;*********************

	.org	0x0000
	rjmp	setup		; Reset Handle
.if	UART_SW==TRUE
	.org	OVF0addr
	rjmp	tim0_ovf	; Timer0 Overflow Handle
.endif


;***************************************************************************
;*
;* INTERRUPT
;*	tim0_ovf 
;* 
;* DESCRIPTION
;* 	Software UART Service Routine. The ISR is only enabled within a
;*      the polling loop, therefor threre is no need to save the status 
;*      register. 
;*
;***************************************************************************

.if 	UART_SW==TRUE
tim0_ovf:
	ldi	temp1, (256-N+10/C);
	;in	temp1, TCNT0
	;subi	temp1, N-3/C
	out	TCNT0, temp1	; TCNT0=baudrate; // reload timer	
	inc	temp2		; u_bitcnt++;
	brtc	to_transmit	; if (t==0) goto transmit

to_receive:
	cpi	temp2, 8	; if (u_bitcnt == 8)
	brne	to_delay	; {
	clr	temp1		;    disable T/C0 Overflow Interrupt
	out	TIMSK, temp1	;
	clt			;    t=0, set receive complete		; }
to_recbit:			;
	lsr	u_data		; u_data=u_data<<1
	sbic	UART_PIN,RXD	; if(UART_PIN.RXD); // (sampled at int+12 cycles) 
	ori	u_data, 0x80	;   u_data=u_data | 0x80;
to_end:				;
	reti			; 

to_delay: 
	rjmp 	to_recbit

to_transmit:			
	cpi	temp2, 10	; if (u_bitcnt == 10)	
	brne	to_2		; {
	sec			;	bit=1;		// generate stop bit
	clr	temp1		;	disable TC0 overflow interrupt
	out	TIMSK, temp1	;
	set			;	t=1; // set transmit complete bit
	rjmp	to_sendbit	;	return;
to_2:	cpi	temp2,1		; else if (u_bitcnt == 1)
	brne	to_3		; {
	clc			;	bit=0;		// generate start bit
	rjmp	to_sendbit	;	return;			; }
to_3:	lsr	u_data		; else { bit=u_data & 1; u_data=u_data>>1; }
	nop			;
to_sendbit:
	brcc	to_4		;
	nop			; if(bit==1)
	sbi	UART_PORT, TXD	;	txd = 1; (set at int+17 cycles)
	rjmp	to_end		; else 
to_4:	cbi	UART_PORT, TXD	;	txd = 0; (set at int+17 cycles)
	rjmp	to_end		; return;
.endif


;***************************************************************************
;*
;* FUNCTION
;*	putc
;*	putc2
;*	putc3
;*	putc4
;*
;* DESCRIPTION
;*	Send one or more characters on the UART Tx line.
;*
;***************************************************************************

putc4:	rcall	putc
	mov	u_data, cmd1	
putc3:	rcall	putc
	mov	u_data, cmd2
putc2:	rcall	putc
	mov	u_data, cmd3
	
putc:	
.if 	UART_SW==TRUE	
	clt			; T=FLAG_TXC=0
 	clr	temp2		; clear bit counter
 	ldi	temp1, 1<<TOV0	; enable T/C0 overflow interrupt
 	out	TIMSK, temp1
putc_wait:	
	brtc	putc_wait
	ret
.else
 	sbis	USR,UDRE
 	rjmp	putc
	out	UDR, u_data
 	ret
.endif	


;***************************************************************************
;*
;* FUNCTION
;*	getc
;*
;* DESCRIPTION
;*	Wait for start bit and receive a character on the UART Rx line.
;*
;***************************************************************************

getc:	
.if	UART_SW==TRUE
	sbis	UART_PIN,RXD	; wait for stop-bit
	rjmp	getc		;
getc0:	
	sbic	UART_PIN,RXD	; <== sample start bit
	rjmp	getc0		;
	ldi	temp1, (256-(N+N/2)+(10+12)/C) 
	out	TCNT0, temp1 	; preset T/C0 to 1.5 bit lengths
	set			; T=FLAG_TXC=1
	ldi	temp1, 1<<TOIE0	;
	out	TIFR, temp1	; clear T/C0 overflow flag
	out	TIMSK, temp1	; enable T/C0 overflow Interrupt
	clr	temp2		; clear bit counter 
getc_wait:
	brts	getc_wait
	ret
.else
	sbis	USR,RXC
	rjmp	getc
	in	u_data, UDR
	ret	
.endif


;***************************************************************************
;*
;* FUNCTION
;*	rdspi 
;*	wrspi 
;*	wrspi_0x00
;*	wrspi_0xac
;*	wrspi_addr
;*	wrspi_addr2
;*	wrspi_u_data
;*	wrspi_pulse_sck
;*
;* DESCRIPTION
;*	 Write and read bytes on the SPI.
;*
;***************************************************************************

wrspi_u_data: 				; called 2 times
	mov	s_data, u_data
	rjmp	wrspi

wrspi_0xac: 				; called 2 times
	ldi	s_data, 0xac	
	rjmp	wrspi

wrspi_wrspi_addr2: 			; called 1 time
	mov	p_cmd  , s_data
	mov	p_addrh, addrh
	mov	p_addrl, addrl
	mov	p_data , u_data

wrspi_wrspi_addr: 			; called 1 time
	rcall	wrspi		

wrspi_addr: 				; never called
	mov	s_data, addrh	
	rcall	wrspi
	mov	s_data, addrl	
	rjmp	wrspi

wrspi_pulse_sck:
	clr	s_data
	ldi	temp1, 1

wrspi_0x00:
	; like rdspi
rdspi:	
	clr	s_data

wrspi:	
	ldi	temp1, 8
wrspi_loop:	
	rol	s_data
	;------ set data output (Programmer applies data) --- 
	brcc	wrspi1
	sbi	SPI_PORT, MOSI
	rjmp	wrspi2
wrspi1:	cbi	SPI_PORT, MOSI
wrspi2:	
	;------ clock low time
	ldi	temp2, CLK/SPI_BPS + 1 
wrspi3:	dec	temp2
	brne	wrspi3
	;------ sample input (Programmer samples data)
	clc
	sbic	SPI_PIN, MISO
	sec
	;------ clock rising edge (AVR stores data)
	sbi	SPI_PORT, SCK
	;------ clock high time
	ldi	temp2, CLK/SPI_BPS + 1
wrspi4:	dec	temp2
	brne	wrspi4
	;------ clock falling edge (AVR applies new data)
	cbi	SPI_PORT, SCK
	;------ next bit
	dec	temp1
	brne	wrspi_loop
	rol	s_data	
	ret


;***************************************************************************
;*
;* FUNCTION
;*	delay
;*	delay_50ms
;*
;* DESCRIPTION
;*	 Make a small delay.
;*
;***************************************************************************

.equ 	tloop  = 4*256-1+3
.equ	T_1ms  = ((( 1*CLK)/1000)+tloop/2)/tloop
.equ	T_2ms  = ((( 2*CLK)/1000)+tloop/2)/tloop
.equ	T_5ms  = ((( 5*CLK)/1000)+tloop/2)/tloop
.equ	T_8ms  = ((( 8*CLK)/1000)+tloop/2)/tloop
.equ	T_10ms = (((10*CLK)/1000)+tloop/2)/tloop
.equ	T_12ms = (((12*CLK)/1000)+tloop/2)/tloop
.equ	T_15ms = (((15*CLK)/1000)+tloop/2)/tloop
.equ	T_20ms = (((20*CLK)/1000)+tloop/2)/tloop
.equ	T_30ms = (((30*CLK)/1000)+tloop/2)/tloop
.equ	T_40ms = (((40*CLK)/1000)+tloop/2)/tloop
.equ	T_50ms = (((50*CLK)/1000)+tloop/2)/tloop

delay_50ms:
	ldi	temp1, T_50ms

delay:	
	clr	temp2	
dl:	dec	temp2	
	nop	
	brne	dl
	dec	temp1
	brne	delay
	ret


;***************************************************************************
;*
;* FUNCTION
;*	inc_addr
;*
;* DESCRIPTION
;*	Auto increment address ****
;*
;***************************************************************************

inc_addr:
	subi	addrl, 0xff
	sbci	addrh, 0xff
	ret


;***************************************************************************
;*
;* FUNCTION
;*	is_page_device
;*
;* DESCRIPTION
;*	sets the zero flag if the current device has page mode capability
;*	in this case temp1 holds the page mask
;*
;***************************************************************************

.equ	page_64  = ~(64-1)
.equ	page_128 = ~(128-1)
.equ	page_256 = ~(256-1)

is_page_device:
	; ================================
	; devices with 64 bytes page size
	; ================================
	ldi	temp1, page_64
	cpi	device, tn26
	breq	is_page_device_return
	cpi	device, m8515	
	breq	is_page_device_return
	cpi	device, m8515B	
	breq	is_page_device_return
	cpi	device, m8535	
	breq	is_page_device_return
	cpi	device, m8535B	
	breq	is_page_device_return
	cpi	device, m8
	breq	is_page_device_return
	cpi	device, m8B
	breq	is_page_device_return
	cpi	device, m83
	breq	is_page_device_return
	cpi	device, m83B
	breq	is_page_device_return
	; ================================
	; devices with 128 bytes page size
	; ================================
	ldi	temp1, page_128
	cpi	device, m16
	breq	is_page_device_return
	cpi	device, m16B
	breq	is_page_device_return
	cpi	device, m161
	breq	is_page_device_return
	cpi	device, m161B
	breq	is_page_device_return
	cpi	device, m163
	breq	is_page_device_return
	cpi	device, m163B
	breq	is_page_device_return
	cpi	device, m32
	breq	is_page_device_return
	cpi	device, m32B
	breq	is_page_device_return
	; ================================
	; devices with 256 bytes page size
	; ================================
	ldi	temp1, page_128
	cpi	device, m64
	breq	is_page_device_return
	cpi	device, m64B
	breq	is_page_device_return
	cpi	device, m128
	breq	is_page_device_return
	cpi	device, m128B
	breq	is_page_device_return
	; =====================================================
	; devices with 256 bytes page size (no polling support)
	; =====================================================
is_old_page_device:	
	cpi	device, m103
	breq	is_page_device_return
	cpi	device, m603
	;breq	is_page_device_return
is_page_device_return:
	ret


;***************************************************************************
;*
;* FUNCTION
;*	cmd_read
;*
;* DESCRIPTION
;*	Perform a AVR command read transfer
;*
;***************************************************************************

cmd_read:
	;rjmp	cmd_write3	;		// alias for cmd_write3


;***************************************************************************
;*
;* FUNCTION
;*	cmd_write3
;*
;* DESCRIPTION
;*	Perform a AVR command write (and read) transfer with byte 4 set to 0
;*
;***************************************************************************

cmd_write3:
	clr	u_data		; u_data=0;	// 4th parameter is zero
	;rjmp	cmd_write
	

;***************************************************************************
;*
;* FUNCTION
;*	cmd_write
;*
;* DESCRIPTION
;*	Perform a AVR command write (and read) transfer
;*
;***************************************************************************

cmd_write:
	mov	s_data, cmd1	; wrspi(cmd1);
	rcall	wrspi		;
	mov	s_data, cmd2	; wrspi(cmd2);
	rcall	wrspi		;
	mov	s_data, cmd3	; wrspi(cmd3);
	rcall	wrspi		;
	mov	s_data, u_data	; return(wrspi(u_data));
	rcall	wrspi
	rjmp	ret_u_data


;***************************************************************************
;*
;* FUNCTION
;*	flash_write_page
;*
;* DESCRIPTION
;*	Write a page in flash memory
;*
;***************************************************************************

flash_write_page:
	ldi	s_data, 0x4c	; wrspi(0x4c);		// SPI write (byte 1)
	rcall	wrspi_wrspi_addr; wrspi(addrh);		// SPI write (byte 2)
				; wrspi(addrl);		// SPI write (byte 3)
	rcall	wrspi_0x00	; wrspi(0x00);		// SPI write (byte 4)
	ldi	cmd1, T_10ms	; better T_14ms to cover 161
	rcall	is_old_page_device
	breq	f_delay
	rjmp	f_poll
	
	
;***************************************************************************
;*
;* FUNCTION
;*	flash_write_lsb
;*
;* DESCRIPTION
;*	Write low byte of flash memory
;*
;***************************************************************************

flash_write_lsb:
	ldi	s_data, 0x40	; wrspi(0x40);		// SPI write (byte 1)
	rjmp	flash_write_byte


;***************************************************************************
;*
;* FUNCTION
;*	flash_write_msb
;*
;* DESCRIPTION
;*	Write high byte of flash memory
;*
;***************************************************************************

flash_write_msb:
	ldi	s_data, 0x48	; wrspi(0x48);		// SPI write (byte 1)

flash_write_byte:
	rcall	wrspi_wrspi_addr2; wrspi(addrh);	// SPI write (byte 2)
				; wrspi(addrl);		// SPI write (byte 3)
	rcall	wrspi_u_data	; wrspi(u_data);	// SPI write (byte 4)
	ldi	cmd1, T_10ms	; default_delay = 10;   // ms
	rcall	is_page_device  ; 
	breq	f_done		;
	cpi	device, S1200D	; if(device!=S1200D) {
	breq	f_delay		;    delay(10; return;				
f_poll: 			; }
	andi	p_cmd, 0x0f	; p_cmd=pcmd & 0x0f;	// generate equivalent
	ori	p_cmd, 0x20	; p_cmd=pcmd | 0x20;    // read command
	
	clr	cmd2		; tries=256;
f_loop: mov	s_data, p_cmd	; do {
	rcall	wrspi		;   wrspi(p_cmd);
	mov	s_data, p_addrh	;   wrspi(p_addrh);
	rcall	wrspi		;
	mov	s_data, p_addrl	;   wrspi(p_addrl);
	rcall	wrspi		;
	rcall	rdspi		;   if(rdspi()==p_data)
	cp	p_data, s_data	;     break;
	breq	f_pass		;
	dec	cmd2		;   tries=tries-1;
	brne	f_loop		; } while(tries!=0);
f_fail:	ldi	u_data, RESP_ERR; putc('?'):
	rjmp	f_ret		; return;
	
f_pass:	tst	cmd2		; if(tries==0) {
	brne	f_done		;
	
f_delay:mov	temp1, cmd1	;   delay(default_delay);
	rcall	delay		; }
	
f_done: ldi	u_data, RESP_OK	; return 0x0d;
f_ret:	cpi	u_data, RESP_ERR;
	ret			;


;***************************************************************************
;*
;* FUNCTION
;*	flash_read_lsb
;*
;* DESCRIPTION
;*	Read byte from data memory
;*
;***************************************************************************

flash_read_lsb:
	ldi	s_data, 0x20	; wrspi(0x20);		// SPI write (byte 1)
.if	SUPPORT_BLOCK_CMD = TRUE	;
	rjmp	read_delay		
.else
	rjmp	read		
.endif

;***************************************************************************
;*
;* FUNCTION
;*	flash_read_msb
;*
;* DESCRIPTION
;*	Read byte from data memory
;*
;***************************************************************************

flash_read_msb:
	ldi	s_data, 0x28	; wrspi(0x28);		// SPI write (byte 1)
.if	SUPPORT_BLOCK_CMD = TRUE; #if SUPPORT_BLOCK_CMD = TRUE
read_delay:			;
	cpi	device, S1200D	;  if(device==S1200D) {
	brne	read		;     
	ldi	temp1, 1	;    delay(0.25);
	rcall	delay		;  }
.endif				; #endif
	rjmp	read		; goto read;


;***************************************************************************
;*
;* FUNCTION
;*	eeprom_write
;*
;* DESCRIPTION
;*	Write byte to data memory
;*
;***************************************************************************

eeprom_write:
	ldi	s_data, 0xc0	; wrspi(0xc0);		// SPI write (byte 1)
	rcall	wrspi_wrspi_addr; wrspi(addrh);		// SPI write (byte 2)
				; wrspi(addrl);		// SPI write (byte 3)
	rcall	wrspi_u_data	; wrspi(u_data);	// SPI write (byte 4)
	ldi	u_data, RESP_OK	;
	ldi	temp1, T_8ms	; delay(8);		// 10 ms fails !?
	rjmp	delay		;
		

;***************************************************************************
;*
;* FUNCTION
;*	eeprom_read
;*
;* DESCRIPTION
;*	Read byte from data memory
;*
;***************************************************************************
	
eeprom_read:	
	ldi	s_data, 0xa0	; wrspi(0xA0);		// SPI write (byte 1)
read:	rcall	wrspi_wrspi_addr; wrspi(addrh);		// SPI write (byte 2)
				; wrspi(addrl);		// SPI write (byte 3)
	rcall	rdspi		; u_data=rdspi();	// Read data (byte 4)
ret_u_data:
	mov	u_data, s_data	; return u_data;
	ret
			
	
;***************************************************************************
;*
;* FUNCTION
;* 	setup
;*
;* DESCRIPTION
;*	Initialization
;*
;***************************************************************************

setup:	
	
	;*******************************************
	;**** Init Stack
	;*******************************************

.ifdef	SPL
	.message "Configuration: software stack enabled"	
	ldi	temp1, low(STACKTOP)
	out	SPL, temp1
.ifdef	SPH
	ldi	temp1, high(STACKTOP)
	out	SPH, temp1
.endif
.else
	.message "Configuration: harware stack enabled"	
.endif	
	
	
	;*******************************************
	;**** Init Ports default value
	;*******************************************
	
	ser	temp1		; Initialize all port pins as ... 
				; ... inputs with pullup rersistors enabled ...
.ifdef	PORTD
	out	PORTD, temp1	
.endif
.ifdef	PORTC
	out	PORTC, temp1	
.endif
.ifdef	PORTB
	out	PORTB, temp1	
.endif
.ifdef	PORTA
	out	PORTA, temp1	; 
.endif
	
	;*******************************************
	;**** Init SPI
	;*******************************************
				
	;sbi	SPI_PORT, RESET ; RESET is inactive (is already default)
	sbi	SPI_DDR, RESET	; RESET is output
	;cbi	SPI_DDR, MISO	; MISO  is input (is already default)
	;cbi	SPI_DDR, MOSI	; MODI  is input (is already default)
	;cbi	SPI_DDR, SCK	; SCK   is input (is already default)

				
	;*******************************************
	;**** Init UART
	;*******************************************			

.if	UART_SW==TRUE		
	.message "Configuration: Software UART enabled"	
	ldi	temp1, R	
	out	TCCR0, temp1
.else
	.message "Configuration: Hardware UART enabled"
	ldi	temp1, BITLEN/16-1
	out	UBRR, temp1
	ldi	temp1,(1<<TXEN) | (1<<RXEN)
	out	UCR, temp1
.endif
	sbi	UART_DDR, TXD	; TXD is output
	;cbi	UART_DDR, RXD	; RXD is input (is already default)


	;*******************************************
	;**** Init LED
	;*******************************************			

.if	LED_USE==TRUE
	cbi	LED_PORT, LED	; LED is off
	sbi	LED_DDR,  LED	; LED is output
.if	LED_DUAL==TRUE
	.message "Configuration: Dual-LED support enabled"
	sbi	LED_PORT, LEDH	; LEDH is on
	sbi	LED_DDR,  LEDH	; LEDH is output
.endif
.endif


	;*******************************************
	;**** Init Application
	;*******************************************
	
	ldi	device, S2313A
	sei			; Enable interrupts


	;*******************************************
	;**** Healtcheck
	;*******************************************

.if	SUPPORT_HEALTH_CHECK = TRUE
.if	LED_USE == TRUE
	.message "Configurtion: Healthcheck support enabled"		;
	;cbi	LED_PORT,LED	; led=0; (is default)
	;sbi	LED_PORT,LEDH	; led=1; (is default)
	rcall	delay_50ms	; delay(50);

	ldi	s_data, 200	;
setup_loop:			; do {
	cbi	LED_PORT,LED	;   led=0;
.if	LED_DUAL==TRUE		;   #if LED_DUAL==TRUE
	sbi	LED_PORT,LEDH	;    ledh=1;
.endif				;   #endif
	ldi	temp1, T_2ms	;   delay(2);
	rcall	delay		;
	sbi	LED_PORT,LED	;   led=1;
.if	LED_DUAL==TRUE		;   #if LED_DUAL==TRUE
	cbi	LED_PORT,LEDH	;    ledh=0;
.endif				;   #endif
	ldi	temp1, T_1ms	;   delay(1);
	rcall	delay		;
	dec	s_data		;   s_data=s_data-1;
	brne	setup_loop	; } while(s_data!=0);
.else
	.error "Error: Cannot enable healthcheck support since no LED is enabled"
.endif
.endif

 	rjmp	waitcmd


;***************************************************************************
;*
;* PROGRAM
;*	waitcmd (main loop)
;*
;* DESCRIPTION
;*	Wait for and execute commands.
;*
;***************************************************************************


	;*******************************************
	;**** Set LED Command
	;*******************************************
w22:	
.if	LED_BUG==TRUE
	.message "Configurtion: LED support ('x' command) bug corrected"		;
	rcall	getc		; getc();
.endif		
.if	LED_USE==TRUE	;
	.message "Configurtion: LED support ('x' command) enabled"		;
	sbi	LED_PORT,LED	; led=1
.if	LED_DUAL==TRUE		; #if LED_DUAL==TRUE
	cbi	LED_PORT,LEDH	;   ledh=0;
.endif				; #endif
.endif
	rjmp	putc_cr		; putc(CR);
		
	
	;*******************************************
	;**** Clr LED Command
	;*******************************************
w23:	
.if	LED_BUG==TRUE
	.message "Configurtion: LED support ('y' command) bug corrected"
	rcall	getc		; getc();
.endif				;
.if	LED_USE==TRUE
	.message "Configurtion: LED support ('y' command) enabled"		;
	cbi	LED_PORT,LED	; led=0;
.if	LED_DUAL==TRUE		; #if LED_DUAL==TRUE
	sbi	LED_PORT,LEDH	;   ledh=0;
.endif				; #endif
.endif
	rjmp	putc_cr		; putc(CR);

	

waitcmd:
	rcall	getc		; u_data=getc();
	
	;=======================================================
	;Command decoder for commands with additonal parameters
	;=======================================================
	cpi	u_data, CMD_ESC			; case ESC:
	breq	waitcmd				;   break;
	cpi	u_data, CMD_DEVICE_TYPE		; case 'T':
	breq	w0      			;   goto w0;
	cpi	u_data, CMD_NEW_UNIVERSAL	; case '.': 
	breq	w2      			;   goto w2
	cpi	u_data, CMD_LOAD_ADDRESS	; case 'A': 
	breq	w3      			;   goto w3
.if	SUPPORT_BYTE_CMD==TRUE 
	cpi	u_data, CMD_WRITE_PROG_HIGH	; case 'C': 
	breq	w4      			;   goto w4
	cpi	u_data, CMD_WRITE_PROG_LOW	; case 'c': 
	breq	w5      			;   goto w5
	cpi	u_data, CMD_WRITE_DATA		; case 'D': 
	breq	w7      			;   goto w7
.endif						; 
	cpi	u_data, CMD_WRITE_LOCK_BITS	; case 'l': 
	breq	w8      			;   goto w8
.if 	SUPPORT_BLOCK_CMD==TRUE			;
	cpi	u_data, CMD_READ_MEMORY		; case 'g':
	breq	w10a				;   goto w10a
	cpi	u_data, CMD_WRITE_MEMORY	; case 'B': 
	breq	w8a      			;   goto w8a
.endif						;
	cpi	u_data, CMD_SET_LED		; case 'x':
	breq	w22				;   goto w22
	cpi	u_data, CMD_CLR_LED     	; case 'y':
	breq	w23				;   goto w23
	rjmp	w9				; default:
						;   goto w9 // all other commands

	;*******************************************
	;**** Device Type 
	;*******************************************
w0:
	rcall	getc		; device=getc();
	mov	device,u_data	; 
	rjmp	putc_cr		; putc(CR);


	;*******************************************
	;**** New Universal Command 
	;*******************************************
w2:	
	rcall	getc		; cmd1 = getc();
	mov	cmd1, u_data	;
	rcall	getc		; cmd2 = getc();
	mov	cmd2, u_data	;
	rcall	getc		; cmd3 = getc();
	mov	cmd3, u_data	;
	rcall	getc		; u_data = getc();
	rcall	cmd_write	; putc(cmd_write());
	rcall	putc		;
	rjmp	delay_50ms_putc_cr; delay(0xFF)
				; putc(CR);
	
	;*******************************************
	;**** Load Address 
	;*******************************************
w3:	
	rcall	getc		; addrh = getc();
	mov	addrh, u_data	;
	rcall	getc		; addrl = getc();
	mov	addrl, u_data	;
	rjmp	putc_cr		; putc(CR);


	;*******************************************
	;**** Write Program Memory, High Byte 
	;*******************************************
w4:	
.if 	SUPPORT_BYTE_CMD==TRUE
	rcall	getc		; u_data = getc();
	rcall	flash_write_msb	;
	rcall	inc_addr;	; addr++; 		// Auto increment address
	rjmp	putc_waitcmd	;
.endif	
	
	;*******************************************
	;**** Write Program Memory, Low Byte 
	;*******************************************
w5:	
.if 	SUPPORT_BYTE_CMD==TRUE
	rcall	getc		; u_data = getc();
	rcall	flash_write_lsb	;
	rjmp	putc_waitcmd	;
.endif	

	;*******************************************
	;**** Write Data Memory 
	;*******************************************
w7:	
.if 	SUPPORT_BYTE_CMD==TRUE
	rcall	getc		; 
	rcall	eeprom_write	; eeprom_write(getc());
	rcall	inc_addr	; addr++;
	rjmp	putc_waitcmd	; putc(CR)
.endif	

	;*******************************************
	;**** Write Lock Bits 
	;*******************************************
w8:	
	rcall	getc		;
	ldi	cmd1, 0xac	;
	mov	cmd2, u_data	;
	andi	cmd2, 0x06	;
	ori	cmd2, 0xe0	;
	ldi	cmd3, 0x00	;
	rcall	cmd_write3	; cmd_write(0xac,getc(),0xe0,0x00);
	rjmp	delay_50ms_putc_cr ; delay(50);
				; putc(CR);
	
	;*******************************************
	;**** Read Memory Block
	;*******************************************
w10a:	
.if 	SUPPORT_BLOCK_CMD==TRUE
	.message "Configurtion: 'g' command enabled"		;
	rcall	getc		; lenh=getc(); 		// msb
	mov	cmd3, u_data	;
	rcall	getc		; lenl=getc(); 		// lsb
	mov	cmd2, u_data	;
	rcall	getc		;
	cpi	u_data, 'E'	; if(getc()=='E')
	breq	w10_loop	;   goto read_eeprom

w10a_loop:			; do {
	rcall	flash_read_lsb	;   putc(flash_read_lsb());
	rcall	putc		;
	rcall	flash_read_msb	;   putc(flash_read_lsb());
	rcall	putc		;
	rcall	inc_addr	;   addr++;		// Auto increment address
	subi	cmd2, 2		;   len-=2;
	sbci	cmd3, 0		;
	brne	w10a_loop	; } while(len!=0);
	rjmp	waitcmd		;
.endif	

	;*******************************************
	;**** Write Memory Block 
	;*******************************************
w8a:
.if 	SUPPORT_BLOCK_CMD==TRUE
	.message "Configurtion: 'B' command enabled"		;
	rcall	getc		; lenh = getc();	// msb
	mov	cmd3, u_data	;
	rcall	getc		; lenl = getc();	// lsb
	mov	cmd2, u_data	;
	rcall	getc		; cmd  = getc();	// transfer type
	mov	cmd1, u_data	;	
	subi	cmd2, 1		;

.if	USE_RAM==TRUE		; #if USE_RAM==TRUE
	ldi	ZH, high(BUFEND);
	ldi	ZL, low(BUFEND) ;  p=(char *)FLASHEND;
	mov	count, cmd2	;  count=len-1
.else				; #else
.ifdef	SPL			;  
	ldi	ZH, 0		;   
.endif				;  
	mov	ZL, cmd2	;  p=len-1;
.endif				; #endif
w_loop:				; do {
	rcall	getc		;    *p = getc();
	st	Z, u_data	;
.if	USE_RAM==TRUE		;    #if USE_RAM==TRUE
	sbiw	ZL,1		;     p=p-1;
	dec	count		;     i=i-1;
.else				;    #else
	dec	ZL		;     p=p-1
.endif				;    #endif;
	brpl	w_loop		; } while(p>=0);

.if	USE_RAM==TRUE		; #if USE_RAM==TRUE
	ldi	ZH, high(BUFEND);
	ldi	ZL, low(BUFEND) ;  p=(char *)FLASHEND;
	mov	count, cmd2	;  count=len-1
.else				; #else
.ifdef	SPL			;
	ldi	ZH, 0		;
.endif				;
	mov	ZL, cmd2	;  p=len-1;
.endif				; #endif

	cpi	cmd1, 'F'	; if(cmd!='F') {
	breq	w_flash_loop	;

w_eeprom_loop:			;    do {
	ld	u_data, Z	;
	rcall	eeprom_write	;	eeprom_write(*p);
	rcall	inc_addr	;	addr++; 
.if 	USE_RAM==TRUE		;       #if USE_RAM==TRUE
	sbiw	ZL,1		;	 p=p-1
	dec	count		;	 i=i-1;
.else				;       #else
	dec	ZL		;	 p=p-1;
.endif				;       #endif
	brpl	w_eeprom_loop	;    } while(p>=0);
	rjmp	putc_waitcmd	;    putc(CR);
				; }
	
w_flash_loop:			; else {
	ld	u_data, Z	;   flash_write_lsb(*p);		
	rcall	flash_write_lsb	
	breq	w_err		;
.if	USE_RAM==TRUE		;   #if USE_RAM==TRUE
	sbiw	ZL,1		;    p=p-1
	dec	count		;    i=i-1
.else				;   #else	;   
	dec	ZL		;    p=p-1;
.endif				;   #endif
	ld	u_data, Z	;   flash_write_msb(*p);	
	rcall	flash_write_msb ;
	breq	w_err		; 
w_new:	rcall	is_page_device	;   if(is_page_device(&temp1)) {
	brne	w_next		; 
	and	temp1, addrl	;      if(temp1 & addr) 
	brne	w_next		;
	rcall	flash_write_page;	  flash_write_page();
	;proceed to next word   ;   }   
w_next:	rcall	inc_addr;	;   addr++; 		// Auto increment address
.if	USE_RAM==TRUE		;   #if USE_RAM==TRUE
	sbiw	ZL,1		;    p=p-1;
	dec	count		;    i=i-1
.else				;   #else
	dec	ZL		;   p=p-1;
.endif
	brpl	w_flash_loop	; } while(p>=0);		;
	rjmp	putc_waitcmd	; putc(CR);
w_err:	rjmp	putc_err	; 
.endif

	;*******************************************
	;**** Write Program Memory Page 
	;*******************************************
w9:	
.if	SUPPORT_BYTE_CMD==TRUE
	cpi	u_data,	CMD_WRITE_PROG_PAGE 
	brne	w10      	; case 'm'
	rcall	flash_write_page; putc(flash_write_page()); 
	rjmp 	putc_waitcmd	; 
.endif

	;*******************************************
	;**** Read Data Memory 
	;*******************************************
w10:	
.if 	SUPPORT_BYTE_CMD==TRUE
	cpi	u_data,CMD_READ_DATA
	brne	w11		; case 'd':
	ldi	cmd3, 0		; len=1;
	ldi	cmd2, 1		;
.else	
	rjmp	w11
.endif	
w10_loop:
	rcall	eeprom_read	; do {
	rcall	putc		;    putc(eeprom_read());
	rcall	inc_addr	;    addr++;
	subi	cmd2,1		;    len=len-1;
	sbci	cmd3,0		;
	brne	w10_loop	; } while(len!=0);
	rjmp	waitcmd		; 
		
			
	;*******************************************
	;**** Read Program Memory
	;*******************************************
w11:	
.if 	SUPPORT_BYTE_CMD==TRUE
	cpi	u_data, CMD_READ_PROG
	brne	w12		; case 'R':
	rcall	flash_read_msb	; putc(flash_read_msb());
	rcall	putc		;
	rcall	flash_read_lsb	; putc(flash_read_lsb());
	rcall	putc		;
	rcall	inc_addr	; addr++;		// Auto increment address
	rjmp	waitcmd		;
.endif
						
	;*******************************************
	;**** Enter Programming Mode 
	;*******************************************
w12:	
	cpi	u_data,	CMD_ENTER_PROG_MODE
	brne	w13		; case 'P':

	sbi	SPI_DDR, MOSI	; MOSI is output
	sbi	SPI_DDR, SCK	; SCK is output
	cbi	SPI_PORT, SCK	; clear SCK
	sbi	SPI_PORT, RESET	; set_reset();
	rcall	delay_50ms	; delay(0xff)
	cbi	SPI_PORT, RESET	; clr_reset();
	rcall	delay_50ms	; delay(0xff)

	rcall	wrspi_0xac	; wrspi(0xac);		// SPI write (byte 1)
	ldi	s_data, 0x53	; wrspi(0x53);		// SPI write (byte 2)
	rcall	wrspi
	cpi	device, 0x20	; if ( (device >= 0x20) && (device <= 0x7F) )
	brlo	s2		;
	tst	device		;
	brmi	s2		; {
	ldi	cmd1, 32	;    count = 32;
s1:				;    do {
	rcall	rdspi		;       if(rdspi()==0x53) // SPI read (byte 3)
	cpi	s_data,0x53
	breq	s3		;	   break;
	rcall	wrspi_0x00	;	wrspi(0x00);	// SPI write (byte 4)
	rcall	wrspi_pulse_sck	;	wrspi_pulse_sck(); // try next bit
	rcall	wrspi_0xac	;	wrspi(0xac);	// SPI write (byte 1)
	ldi	s_data,0x53	;	wrspi(0x53);	// SPI write (byte 2)
	rcall	wrspi		;
	dec	cmd1		;    	--count;
	brne	s1		;    } while(count!=0);
	rjmp	s3		; }
				; else { // SPI Synchronization (fix!)
s2:	rcall	wrspi_0x00	;    wrspi(0x00);	// SPI write (byte 3)
				; }
s3:	rcall 	wrspi_0x00	; wrspi(0x00);		// SPI write (byte 4)
	rjmp	delay_50ms_putc_cr; delay(50);
				; putc(CR);


	;*******************************************
	;**** Leave Programming Mode 
	;*******************************************
w13:	
 	cpi	u_data, CMD_EXIT_UPGRADE
	breq	w131		; case 'E':
	cpi	u_data, CMD_LEAVE_PROG_MODE
	brne	w14		; case 'L':
w131:	cbi	SPI_DDR, MOSI	; MOSI is tri-state
	cbi	SPI_DDR, SCK	; SCK is tri-state
	sbi	SPI_PORT, RESET	; set_reset();
	rjmp    putc_cr		; putc(CR);


	;*******************************************
	;**** Chip Erase 
	;*******************************************
w14:	
	cpi	u_data, CMD_CHIP_ERASE      
        brne	w15		; case 'e':
        ldi	cmd1, 0xac	;
        ldi	cmd2, 0x80	;
        ldi	cmd3, 0x00	;
        rcall	cmd_write3	; cmd_write(0xac,0x80,0x00,0x00);
	rjmp	delay_50ms_putc_cr ; delay(50);
				; putc(CR);
				
	;*******************************************
	;**** Read Signature Bytes 
	;*******************************************
w15:
	cpi	u_data, CMD_READ_SIGNATURE
	brne	w16		; case 's':
	ldi	cmd1, 0x30	;
	ldi	cmd2, 0x00	;
	ldi	cmd3, 0x02	; cmd3=0x02;
w15_loop:			; do {
	rcall	cmd_read	;   putc(cmd_read(0x30,0x00,cmd3));
	rcall	putc		;
	subi	cmd3, 1		;   cmd3--;
	brpl	w15_loop	; } while(cmd3>=0);	
	rjmp	waitcmd		;
	
	
	;*******************************************
	;**** Return Software Identifier 
	;*******************************************
w16:	
	cpi	u_data, CMD_RETURN_SW_ID
	brne	w17		; case 'S': 
	ldi	u_data, 'A'	; putc('A');
	ldi	cmd1  , 'V'	; putc('V');
	ldi	cmd2  , 'R'	; putc('R');
	ldi	cmd3  , ' '	; putc(' ');
	rcall	putc4		;
	ldi	u_data, 'I'	; putc('I');
	ldi	cmd2  , 'S'	; putc('S');
	ldi	cmd3  , 'P'	; putc('P');
w161:	rcall   putc3		;
	rjmp	waitcmd		;
		

	;*******************************************
	;**** Return Software Version 
	;*******************************************
w17:	
	cpi	u_data, CMD_RETURN_SW_VERSION
	brne	w18		; case 'V': 
	ldi	u_data, SW_MAJOR; putc(SW_MAYOR);
	ldi	cmd3  , SW_MINOR; putc(SW_MINOR);
w171:	rcall	putc2		;
	rjmp	waitcmd		;


	;*******************************************
	;**** Return Hardware Version
	;*******************************************
w18:	
.if	SUPPORT_v_CMD==TRUE
	.message "Configurtion: 'v' command enabled"
	cpi	u_data, CMD_RETURN_HW_VERSION
	brne	w19		; case 'v':
	ldi	u_data, HW_MAJOR; putc(HW_MAJOR);
	ldi	cmd3  , HW_MINOR; putc(HW_MINOR);
	rjmp	w171		;
.endif


	;*******************************************
	;**** Return Programmer Type 
	;*******************************************
w19:	
	cpi	u_data, CMD_RETURN_PROGRAMMER_TYPE
	brne	w20		; case 'p':
	ldi	u_data,0x53     ; putc('S');	// serial programmer
	rjmp	putc_waitcmd	;


	;*******************************************
	;**** Return autoincrement address support
	;*******************************************
w20:	
	cpi	u_data, CMD_RETURN_AUTOINC_SUPPORT
	brne	w20a		; case 'a':
	ldi	u_data,'Y'	; putc('Y'); 	
	rjmp	putc_waitcmd	;


	;*******************************************
	;**** Return block write support
	;*******************************************
w20a:	
.if 	SUPPORT_BLOCK_CMD==TRUE
	.message "Configurtion: 'b' command enabled"		;
	cpi	u_data, CMD_RETURN_BLOCK_WRITE_SUPPORT
	brne	w21			; case 'b': 
	ldi	u_data, 'Y'		; putc('Y'); 	
	ldi	cmd2, high(UART_RXSIZE) ; putc((BUFSIUE>>8) & 0xff);
	ldi	cmd3, low (UART_RXSIZE)	; putc(BUFSIUE&0xff);
	rjmp	w161		;
.endif	

	;*******************************************
	;**** Show Supported Devices 
	;*******************************************
w21:	
	cpi	u_data, CMD_RETURN_SUPPORTED_DEVICES
	brne	w99		; case 't':
	ldi	cmd1, 0x7f	; dev=0x7f;
w21_loop:
	mov	u_data, cmd1	; do {
	rcall	putc		;   putc(dev);
	dec	cmd1		;   dev=dev-1;
	brpl	w21_loop	; } while(dev>=0);
	rjmp	waitcmd		;
	

	;*******************************************
	;**** Unknown Command 
	;*******************************************
w99:				; default:
	;rjmp	putc_err	; putc('?');
	
	
	;*******************************************
	;**** Command Error 
	;*******************************************
	
putc_err:
	ldi	u_data,RESP_INVALID_CMD	; putc('?');	
	rjmp	putc_waitcmd		;


	;*******************************************
	;**** Command OK 
	;*******************************************
	
delay_50ms_putc_cr:
	rcall	delay_50ms	; delay(50);

putc_cr:
	ldi	u_data, RESP_OK	; putc(CR);

putc_waitcmd:	
	rcall	putc		;
	rjmp	waitcmd		;

	
	;.macro	pdef
	;.equ	@0_PORT=@1
	;.equ	@0_DDR =@1
	;.equ	@0_PIN =@1
	;.endm

	;.macro	pset
	;	cbi	@0_PORT,@0	
	;.endm

	;.macro	pclr
	;	cbi	@0_PORT,@0	
	;.endm


	;.listmac
	;pdef	LEDH,1
	;pset	LED
	;cbi	LED_PORT,LED

	;*******************************************
	;**** Chip ID
	;*******************************************

.if	SUPPORT_CHIP_ID==TRUE
	.message "Configurtion: chip ID enabled"		;
	.org	FLASHEND
	.dw	(HW_MAJOR-'0')<<12 | (HW_MINOR-'0')<<8 | (SW_MAJOR-'0')<<4 | (SW_MINOR-'0')<<0
.endif	
